##
# File: DetachedProcessBase.py
# Date: 10-Feb-2015  J. Westbrook --
#
# Update: 4-Mar-2016 jdw add explicit adjustments to the umask - if it appears that
#                        unset -- (e.g. default assignment 022).
#
##

import atexit
import os
import sys
import time
from signal import SIGKILL
from signal import SIGTERM

import psutil


class DetachedProcessBase(object):

    """
    Base class for managing a detached process.

    Subclass the base class run() method as desired.
    """

    def __init__(self, pidFile="/tmp/DetachedProcessBase.pid", stdin=os.devnull, stdout=os.devnull, stderr=os.devnull, wrkDir="/", gid=None, uid=None):
        self.__stdin = stdin
        self.__stdout = stdout
        self.__stderr = stderr
        self.__pidFile = pidFile
        self.__wrkDir = wrkDir
        if uid is None:
            uid = os.getuid()
        self.__uid = uid
        if gid is None:
            gid = os.getgid()
        self.__gid = gid

    def __detachPrep(self):
        """
        Internal method to prepare the execution environment for the detached process.

        """
        # Fork twice and then redirect the standard file descriptors -
        #
        # First fork -
        try:
            pid = os.fork()
            if pid > 0:
                # clean exit from the first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("+DetachedProcessBase.__detachPrep(): Failing with %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        #
        os.chdir(self.__wrkDir)
        #
        # Certain privileges will be required to change owner and group -
        self.__setOwnerGroup(self.__uid, self.__gid)
        #
        os.setsid()
        #
        #  Test for a sensible umask -
        #
        utest = os.umask(0o022)
        if utest < 2:
            os.umask(0o022)
        else:
            os.umask(utest)
        #
        #  Second fork -
        try:
            pid = os.fork()
            if pid > 0:
                # clean exit from the second parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("+DetachedProcessBase.__detachPrep(): Failing with %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # Redirect the stdin, stdout and stderr descriptors to the alternative files assigned in the
        #  constructor --
        sys.stdout.flush()
        sys.stderr.flush()
        stdInFh = open(self.__stdin, "r")
        os.dup2(stdInFh.fileno(), sys.stdin.fileno())
        #

        if sys.version_info[0] > 2:
            # Unbuffered text i/o not allowed - binary mode - which is fine for logging
            stdOutFh = open(self.__stdout, "ab+", 0)
            stdErrFh = open(self.__stderr, "ab+", 0)
        else:
            stdOutFh = open(self.__stdout, "a+", 0)
            stdErrFh = open(self.__stderr, "a+", 0)
        os.dup2(stdOutFh.fileno(), sys.stdout.fileno())
        os.dup2(stdErrFh.fileno(), sys.stderr.fileno())

        # Register an exit handler to cleanup the process id file -
        atexit.register(self._deletePidFile)

        # Store the process id for the detached process -
        pid = str(os.getpid())
        with open(self.__pidFile, "w+") as fout:
            fout.write("%s\n" % pid)

    def _deletePidFile(self):
        """
        Method to remove the sentinnel file containing the process id.
        """
        if os.path.exists(self.__pidFile):
            os.remove(self.__pidFile)

    def __getPidFromFile(self):
        """
        Internal method to read process id file and return the process id.
        """
        try:
            pf = open(self.__pidFile, "r")
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None
        return pid

    def __isRunning(self):
        """Internal method to read the current process id and check if the process
        is active.
        """
        try:
            # This old API was obsoleted with psutil 2.0.0
            pid = self.__getPidFromFile()
            #  Further check if the process id is active -
            if pid in psutil.get_pid_list():  # pylint: disable=no-member
                return True
            else:
                return False
        except Exception as e:  # noqa: F841  pylint: disable=unused-variable
            try:
                pid = self.__getPidFromFile()
                #  Further check if the process id is active -
                if pid in psutil.pids():
                    return True
                else:
                    return False

            except Exception as _e:  # noqa: F841
                pass
        return False

    def __setOwnerGroup(self, uid, gid):
        """
        Internal method to set the owner UID and GID of this process.  Requires a special privileges
        to change the uid/gid.
        """
        try:
            os.setgid(gid)
            os.setuid(uid)
            return True
        except Exception as e:
            sys.stderr.write("+DetachedProcessBase.__setOwnerGroup failing (%s)\n" % str(e))
        return False

    def start(self):
        """
        Start the application as detached process -
        """
        if self.__isRunning():
            sys.stderr.write("+DetachedProcessBase.start(): Process file %s exists and process is running.\n" % self.__pidFile)
            sys.exit(1)

        self.__detachPrep()
        self.run()

    def stop(self):
        """
        Stop the detached process and remove the process id file -

        Kill current process and any descend/child processes.

        """
        # Graceful suspend activity prior to the kill -
        self.suspend()
        #
        pid = self.__getPidFromFile()
        if not pid:
            sys.stderr.write("+DetachedProcessBase.stop(): Process file %s does not exist.\n" % self.__pidFile)
            return

        try:
            p = psutil.Process(pid)
            cPidList = p.children(recursive=True)
            for cPid in cPidList:
                iPid = int(str(cPid.pid))
                os.kill(iPid, SIGTERM)

            while True:
                os.killpg(os.getpgid(pid), SIGKILL)
                os.kill(pid, SIGKILL)
                time.sleep(0.1)
        except Exception as err:
            err = str(err)
            if (err.find("No such process") != -1) or (err.find("no process found") != -1):
                self._deletePidFile()
            else:
                sys.stderr.write(str(err))
                sys.exit(1)

    def restart(self):
        """
        Restart the application as a detached process --
        """
        self.stop()
        self.start()

    def status(self):
        """
        Report the current status of the detached process and any process descendents.
        """
        msgList = []
        if self.__isRunning():
            pid = self.__getPidFromFile()
            msgList.append("+DetachedProcessBase.status(): active process id is %d (process group %d)\n" % (pid, os.getpgid(pid)))
            p = psutil.Process(pid)
            cPidList = p.children(recursive=True)
            for cPid in cPidList:
                iPid = int(str(cPid.pid))
                cp = psutil.Process(iPid)
                msgList.append("+DetachedProcessBase.status(): child process id: %d  parent %d name %s (process group %d)\n" % (iPid, cp.ppid(), cp.name(), os.getpgid(pid)))

        else:
            msgList.append("+DetachedProcessBase.status(): No active process is running.\n")

        return "".join(msgList)

    def run(self):
        """
        This is the entry point for the detached process.  Subclass this method and use the
        start(), stop(), restart() and status() methods to manage the process.
        """

    def suspend(self):
        """
        This is an optional entry point to gracefully suspend the detached process before stopping/killing.
        Subclass this method in the worker class. This method will be called prior to the stop() method.
        """
        return True



##
# File: SubProcessUtil.py
# Date: 10-Feb-2015  J. Westbrook
# #
import os
import subprocess
import sys

# import datetime
# import signal
# import stat
# import time


class SubProcessUtil:
    """Skeleton methods supporting running shell and python scripts as subprocesses.

    These methods are provided primarily to support testing other classes.
    """

    def __init__(self, verbose=True, log=sys.stdout):  # pylint: disable=unused-argument
        self.__lfh = log
        # self.__wrkPath = "."

    def runPythonDetached(self, pythonFilePath, arguments="", logFilePath="testlog.log"):
        return self.__runPyDetached(pythonFilePath=pythonFilePath, arguments=arguments, logFilePath=logFilePath)

    def __runPyDetached(self, pythonFilePath, arguments="", logFilePath="testlog.log"):
        """ """
        commandString = "%s %s %s >> %s 2>&1" % (sys.executable, pythonFilePath, arguments, logFilePath)
        return self.__runCommandDetached(commandString)

    def __runCommandDetached(self, commandString):
        self.__lfh.write("SubProcessUtil.__runCommandDetached() running command string:\n %r\n" % commandString)
        pid = subprocess.Popen(  # pylint: disable=subprocess-popen-preexec-fn
            commandString, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, close_fds=True, preexec_fn=os.setsid
        ).pid
        # pid = subprocess.Popen(commandString, stdout=None, stderr=None, shell=True, preexec_fn=os.setsid).pid
        return pid

    # def __runCommandFileDetached(self, commandFilePath):
    #     self.__lfh.write("SubProcessUtil.__runProcessDetached() running command file %r\n" % commandFilePath)
    #     process = subprocess.Popen(  # pylint: disable=subprocess-popen-preexec-fn
    #         commandFilePath, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False, close_fds=True, preexec_fn=os.setsid
    #     )
    #     return process.pid

    # def __wrapInShell(self, commandFilePath, commandString):
    #     """Embed the input command string within a Bourne shell packaged in the
    #     input command file path.
    #     """
    #     try:
    #         ofh = open(commandFilePath, "w")
    #         ofh.write("#!/bin/sh\n")
    #         ofh.write(commandString)
    #         ofh.write("\n#\n")
    #         ofh.close()
    #         st = os.stat(commandFilePath)
    #         os.chmod(commandFilePath, st.st_mode | stat.S_IEXEC)
    #         return True
    #     except Exception as e:  # noqa: F841 pylint: disable=unused-variable
    #         return False

    # def __runPyDetachedInShell(self, pythonFilePath, arguments="", stdoutFilePath=os.devnull, stderrFilePath=os.devnull):
    #     """ """
    #     commandString = "python %s %s 1> %s 2> %s &" % (pythonFilePath, arguments, stdoutFilePath, stderrFilePath)
    #     ok = self.__wrapInShell("./test.sh", commandString)
    #     if ok:
    #         return self.__runCommandFileDetached("./test.sh")
    #     else:
    #         return -1

    # def __runTimeout(self, commandString, timeout, logPath=None):
    #     """Execute the input command string (sh semantics) as a subprocess with a timeout."""
    #     start = datetime.datetime.now()
    #     cmdfile = os.path.join(self.__wrkPath, "timeoutscript.sh")
    #     self.__wrapInShell(commandString, cmdfile)
    #     self.__lfh.write("+__runTimeout() running command %r\n" % cmdfile)
    #     process = subprocess.Popen(  # pylint: disable=subprocess-popen-preexec-fn
    #         cmdfile, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False, close_fds=True, preexec_fn=os.setsid
    #     )
    #     while process.poll() is None:
    #         time.sleep(0.1)
    #         now = datetime.datetime.now()
    #         if (now - start).seconds > timeout:
    #             # os.kill(-process.pid, signal.SIGKILL)
    #             os.killpg(process.pid, signal.SIGKILL)
    #             os.waitpid(-1, os.WNOHANG)
    #             self.__lfh.write("+ERROR __runTimeout() - Execution terminated by timeout %d (seconds)\n" % timeout)
    #             if logPath is not None:
    #                 ofh = open(logPath, "a")
    #                 ofh.write("+ERROR __runTimeout() Execution terminated by timeout %d (seconds)\n" % timeout)
    #                 ofh.close()
    #             return None
    #     self.__lfh.write("+__runTimeout() completed with return code %r\n" % process.stdout.read())
    #     return 0


if __name__ == "__main__":  # pragma: no cover
    spu = SubProcessUtil(verbose=True, log=sys.stdout)
    lpid = spu.runPythonDetached(pythonFilePath="testTask.py", arguments=" --logfile mylog.log", logFilePath="mylog.log")
    sys.stdout.write("PROCESS ID = %d\n" % lpid)
    #
    #



##
# File:    DetachUtils.py
# Date:    06-Mar-2013
#
# Updates:
#   07-Mar-2013  jdw Define method setLogHandle() to implement log redirection.  This method must
#                    must exist on the input object.
##
"""
Manage detached processes

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os
import sys
import time
import traceback


class DetachUtils(dict):
    """Derived dictionary class supporting automatic initialization.

    This will support pickle serialization/deserialization.
    """

    def __init__(self, reqObj=None, verbose=True, log=sys.stderr):  # pylint: disable=super-init-not-called
        self.__verbose = verbose
        self.__lfh = log
        self.__reqObj = reqObj
        self.__sessionObj = self.__reqObj.getSessionObj()
        self.__sessionPath = self.__sessionObj.getPath()

    def set(self, workerObj=None, workerMethod=None):
        try:
            self.__logFunc = getattr(workerObj, "setLogHandle")  # pylint: disable=attribute-defined-outside-init
            self.__workerFunc = getattr(workerObj, workerMethod)  # pylint: disable=attribute-defined-outside-init
            return True
        except AttributeError:
            self.__lfh.write("+DetachUtils.set() object/attribute error\n")
            return False

    def runDetach(self):
        """
        Run the worker function as a detached process --
        """
        #
        siteId = self.__reqObj.getValue("WWPDB_SITE_ID")
        if self.__verbose:
            self.__lfh.write("+DetachUtils.__runDetach() - STARTING\n")

        sph = self.__setSemaphore()
        child_pid = os.fork()
        if child_pid == 0:
            os.setsid()
            sub_pid = os.fork()
            if sub_pid:
                # Parent of second fork
                os._exit(0)  # pylint: disable=protected-access

            sys.stdout = RedirectDevice()
            sys.stderr = RedirectDevice()
            os.setpgrp()
            os.umask(0)
            os.environ["WWPDB_SITE_ID"] = siteId
            #
            # Redirect the logfile for the child process -
            #
            self.__openSemaphoreLog(sph)
            sys.stdout = self.__cLog
            sys.stderr = self.__cLog
            #
            try:
                self.__logFunc(log=self.__cLog)

                if self.__verbose:
                    self.__cLog.write("+DetachUtils.__runDetach() Child Process: PID# %s\n" % os.getpid())
                    self.__cLog.write("+DetachUtils.__runDetach() Site id       %s\n" % siteId)
                #
                ok = self.__workerFunc()

                if ok:
                    self.__postSemaphore(sph, "OK")
                else:
                    self.__postSemaphore(sph, "FAIL")
                self.__cLog.flush()
            except Exception as e:  # noqa: F841 pylint: disable=unused-variable
                traceback.print_exc(file=self.__cLog)
                self.__cLog.write("+DetachUtils.__runDetach() Failing for child Process: PID# %s\n" % os.getpid())
                self.__postSemaphore(sph, "FAIL")
                self.__cLog.flush()
            os._exit(0)  # pylint: disable=protected-access

        else:
            # Parent returns status information only -
            #
            if self.__verbose:
                self.__lfh.write("+DetachUtils.__runDetach() PARENT COMPLETED parent process: pid# %s\n" % os.getpid())
            os.waitpid(child_pid, 0)
            return True

    def semaphoreExists(self, semaphore="TMP_"):
        fPathAbs = os.path.join(self.__sessionPath, semaphore)
        if os.access(fPathAbs, os.F_OK):
            return True
        else:
            return False

    def getSemaphore(self, semaphore="TMP_"):
        fPathAbs = os.path.join(self.__sessionPath, semaphore)
        try:
            fp = open(fPathAbs, "r")
            lines = fp.readlines()
            fp.close()
            sval = lines[0][:-1]
        except Exception as e:  # noqa: F841 pylint: disable=unused-variable
            sval = "FAIL"

        if self.__verbose:
            self.__lfh.write("+DetachUtils.__getSemaphore() - checked %s in path %s returning %s \n" % (semaphore, fPathAbs, sval))

        return sval

    def __setSemaphore(self):
        sVal = str(time.strftime("TMP_%Y%m%d%H%M%S", time.localtime()))
        self.__reqObj.setValue("semaphore", sVal)
        return sVal

    def __openSemaphoreLog(self, semaphore="TMP_"):
        fPathAbs = os.path.join(self.__sessionPath, semaphore + ".log")
        self.__cLog = open(fPathAbs, "w")  # pylint: disable=attribute-defined-outside-init

    # This should be invoked in __del__ likely
    # def __closeSemaphoreLog(self, semaphore="TMP_"):  # pylint: disable=unused-argument
    #     self.__cLog.flush()
    #     self.__cLog.close()

    def __postSemaphore(self, semaphore="TMP_", value="OK"):
        fPathAbs = os.path.join(self.__sessionPath, semaphore)
        fp = open(fPathAbs, "w")
        fp.write("%s\n" % value)
        fp.close()
        return semaphore


class RedirectDevice:
    def write(self, s):
        pass
